# Copyright (C) 2009 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import BaseHTTPServer
import errno
import os
import re
import urlparse
import socket
import spydaap
from spydaap.daap import do


def makeDAAPHandlerClass(server_name, cache, md_cache, container_cache):
    session_id = 1

    class DAAPHandler(BaseHTTPServer.BaseHTTPRequestHandler):
        daap_server_revision = 1
        protocol_version = "HTTP/1.1"

        def h(self, data, **kwargs):
            self.send_response(kwargs.get('status', 200))
            self.send_header('Content-Type', kwargs.get('type', 'application/x-dmap-tagged'))
            self.send_header('DAAP-Server', 'Simple')
            self.send_header('Expires', '-1')
            self.send_header('Cache-Control', 'no-cache')
            self.send_header('Accept-Ranges', 'bytes')
            self.send_header('Content-Language', 'en_us')
            if 'extra_headers' in kwargs:
                for k, v in kwargs['extra_headers'].iteritems():
                    self.send_header(k, v)
            try:
                if isinstance(data, file):
                    self.send_header("Content-Length", str(os.stat(data.name).st_size))
                else:
                    self.send_header("Content-Length", len(data))
            except Exception:
                pass
            self.end_headers()
            if hasattr(self, 'isHEAD') and self.isHEAD:
                pass
            else:
                try:
                    if (hasattr(data, 'next')):
                        for d in data:
                            self.wfile.write(d)
                    else:
                        self.wfile.write(data)
                except socket.error, (err_no, err_str):
                    if err_no in [errno.ECONNRESET]:
                        # XXX: why do we need to pass this?
                        pass
                    else:
                        raise
            if (hasattr(data, 'close')):
                data.close()

        # itunes sends request for:
        # GET daap://192.168.1.4:3689/databases/1/items/626.mp3?seesion-id=1
        # so we must hack the urls; annoying.
        itunes_re = '^(?://[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}:[0-9]+)?'
        drop_q = '(?:\\?.*)?$'

        def do_GET(self):
            parsed_path = urlparse.urlparse(self.path).path
            if re.match(self.itunes_re + "/$", parsed_path):
                self.do_GET_login()
            elif re.match(self.itunes_re + '/server-info$', parsed_path):
                self.do_GET_server_info()
            elif re.match(self.itunes_re + '/content-codes$', parsed_path):
                self.do_GET_content_codes()
            elif re.match(self.itunes_re + '/databases$', parsed_path):
                self.do_GET_database_list()
            elif re.match(self.itunes_re + '/databases/([0-9]+)/items$', parsed_path):
                md = re.match(self.itunes_re + '/databases/([0-9]+)/items$', parsed_path)
                self.do_GET_item_list(md.group(1))
            elif re.match(self.itunes_re + '/databases/([0-9]+)/items/([0-9]+)\\.([0-9a-z]+)' + self.drop_q, parsed_path):
                md = re.match(self.itunes_re + '/databases/([0-9]+)/items/([0-9]+)\\.([0-9a-z]+)' + self.drop_q, parsed_path)
                self.do_GET_item(md.group(1), md.group(2), md.group(3))
            elif re.match(self.itunes_re + '/databases/([0-9]+)/containers$', parsed_path):
                md = re.match(self.itunes_re + '/databases/([0-9]+)/containers$', parsed_path)
                self.do_GET_container_list(md.group(1))
            elif re.match(self.itunes_re + '/databases/([0-9]+)/containers/([0-9]+)/items$', parsed_path):
                md = re.match(self.itunes_re + '/databases/([0-9]+)/containers/([0-9]+)/items$', parsed_path)
                self.do_GET_container_item_list(md.group(1), md.group(2))
            elif re.match('^/login$', parsed_path):
                self.do_GET_login()
            elif re.match('^/logout$', parsed_path):
                self.do_GET_logout()
            elif re.match('^/update$', parsed_path):
                self.do_GET_update()
            else:
                self.send_error(404)
            return

        def do_HEAD(self):
            self.isHEAD = True
            self.do_GET()

        def do_GET_login(self):
            mlog = do('dmap.loginresponse',
                      [do('dmap.status', 200),
                       do('dmap.sessionid', session_id)])
            self.h(mlog.encode())

        def do_GET_logout(self):
            self.send_response(204)
            self.end_headers()

        def do_GET_server_info(self):
            msrv = do('dmap.serverinforesponse',
                      [do('dmap.status', 200),
                       do('dmap.protocolversion', '2.0'),
                       do('daap.protocolversion', '3.0'),
                       do('dmap.timeoutinterval', 1800),
                       do('dmap.itemname', server_name),
                       do('dmap.loginrequired', 0),
                       do('dmap.authenticationmethod', 0),
                       do('dmap.supportsextensions', 0),
                       do('dmap.supportsindex', 0),
                       do('dmap.supportsbrowse', 0),
                       do('dmap.supportsquery', 0),
                       do('dmap.supportspersistentids', 0),
                       do('dmap.databasescount', 1),
                       #do('dmap.supportsautologout', 0),
                       #do('dmap.supportsupdate', 0),
                       #do('dmap.supportsresolve', 0),
                       ])
            self.h(msrv.encode())

        def do_GET_content_codes(self):
            children = [do('dmap.status', 200)]
            for code in spydaap.daap.dmapCodeTypes.keys():
                (name, dtype) = spydaap.daap.dmapCodeTypes[code]
                d = do('dmap.dictionary',
                       [do('dmap.contentcodesnumber', code),
                        do('dmap.contentcodesname', name),
                        do('dmap.contentcodestype',
                            spydaap.daap.dmapReverseDataTypes[dtype])
                        ])
                children.append(d)
            mccr = do('dmap.contentcodesresponse',
                      children)
            self.h(mccr.encode())

        def do_GET_database_list(self):
            d = do('daap.serverdatabases',
                   [do('dmap.status', 200),
                    do('dmap.updatetype', 0),
                    do('dmap.specifiedtotalcount', 1),
                    do('dmap.returnedcount', 1),
                    do('dmap.listing',
                        [do('dmap.listingitem',
                            [do('dmap.itemid', 1),
                             do('dmap.persistentid', 1),
                             do('dmap.itemname', server_name),
                             do('dmap.itemcount',
                                len(md_cache)),
                             do('dmap.containercount', len(container_cache))])
                         ])
                    ])
            self.h(d.encode())

        def do_GET_item_list(self, database_id):
            def build_item(md):
                return do('dmap.listingitem',
                          [do('dmap.itemkind', 2),
                           do('dmap.containeritemid', md.id),
                           do('dmap.itemid', md.id),
                           md.get_dmap_raw()
                           ])

            def build(f):
                children = [build_item(md) for md in md_cache]
                file_count = len(children)
                d = do('daap.databasesongs',
                       [do('dmap.status', 200),
                        do('dmap.updatetype', 0),
                        do('dmap.specifiedtotalcount', file_count),
                        do('dmap.returnedcount', file_count),
                        do('dmap.listing',
                            children)])
                f.write(d.encode())

            data = cache.get('item_list', build)
            self.h(data)

        def do_GET_update(self):
            mupd = do('dmap.updateresponse',
                      [do('dmap.status', 200),
                       do('dmap.serverrevision', self.daap_server_revision),
                       ])
            self.h(mupd.encode())

        def do_GET_item(self, database, item, format):
            try:
                fn = md_cache.get_item_by_id(item).get_original_filename()
            except IndexError:          # if the track isn't in the DB, we get an exception
                self.send_error(404)    # this can be caused by left overs from previous sessions
                return

            if ('Range' in self.headers):
                rs = self.headers['Range']
                m = re.compile('bytes=([0-9]+)-([0-9]+)?').match(rs)
                (start, end) = m.groups()
                if end is not None:
                    end = int(end)
                else:
                    end = os.stat(fn).st_size
                start = int(start)
                f = spydaap.ContentRangeFile(fn, open(fn), start, end)
                extra_headers = {"Content-Range": "bytes %s-%s/%s" % (str(start), str(end), str(os.stat(fn).st_size))}
                status = 206
            else:
                f = open(fn)
                extra_headers = {}
                status = 200
            # this is ugly, very wrong.
            type = "audio/%s" % (os.path.splitext(fn)[1])
            self.h(f, type=type, status=status, extra_headers=extra_headers)

        def do_GET_container_list(self, database):
            container_do = []
            for i, c in enumerate(container_cache):
                d = [do('dmap.itemid', i + 1),
                     do('dmap.itemcount', len(c)),
                     do('dmap.containeritemid', i + 1),
                     do('dmap.itemname', c.get_name())]
                if c.get_name() == 'Library':  # this should be better
                    d.append(do('daap.baseplaylist', 1))
                else:
                    d.append(do('com.apple.itunes.smart-playlist', 1))
                container_do.append(do('dmap.listingitem', d))
            d = do('daap.databaseplaylists',
                   [do('dmap.status', 200),
                    do('dmap.updatetype', 0),
                    do('dmap.specifiedtotalcount', len(container_do)),
                    do('dmap.returnedcount', len(container_do)),
                    do('dmap.listing',
                        container_do)
                    ])
            self.h(d.encode())

        def do_GET_container_item_list(self, database_id, container_id):
            container = container_cache.get_item_by_id(container_id)
            self.h(container.get_daap_raw())

    return DAAPHandler



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import mutagen.id3
import mutagen.mp3
import spydaap
import re
import os
import struct
from spydaap.daap import do
mutagen.id3.ID3.PEDANTIC = False


class Mp3Parser(spydaap.parser.Parser):
    mp3_string_map = {
        'TIT1': 'daap.songgrouping',
        'TIT2': 'dmap.itemname',
        'TPE1': 'daap.songartist',
        'TPE3': 'daap.songalbumartist',
        'TCOM': 'daap.songcomposer',
        'TCON': 'daap.songgenre',
        'TALB': 'daap.songalbum',
    }

    mp3_int_map = {
        'TBPM': 'daap.songbeatsperminute',
        'TDRC': 'daap.songyear',
        'TCMP': 'daap.songcompilation',
        #'TLEN': 'daap.songtime',
    }
#do('daap.songdiscnumber', 1),
#        do('daap.songgenre', 'Test'),
#        do('daap.songdisccount', 1),
#        do('daap.songcompilation', False),
#        do('daap.songuserrating', 1),

    def handle_rating(self, mp3, d):
        popm = mp3.tags.getall('POPM')
        if popm is not None and len(popm) > 0:
            rating = int(popm[0].rating * (0.39215686274509803))
            d.append(do('daap.songuserrating', rating))

    def handle_track(self, mp3, d):
        tracknumber = None
        trackcount = None
        if 'TRCK' in mp3.tags:
            t = str(mp3.tags['TRCK']).split('/')
            tracknumber = self.my_int(t[0])
            if (len(t) == 2):
                trackcount = self.my_int(t[1])
        if tracknumber:
            d.append(do('daap.songtracknumber', tracknumber))
        if trackcount:
            d.append(do('daap.songtrackcount', trackcount))

    def handle_disc(self, mp3, d):
        discnumber = None
        disccount = None
        if 'TPOS' in mp3.tags:
            t = str(mp3.tags['TPOS']).split('/')
            discnumber = self.my_int(t[0])
            if (len(t) == 2):
                disccount = self.my_int(t[1])
        if discnumber:
            d.append(do('daap.songdiscnumber', discnumber))
        if disccount:
            d.append(do('daap.songdisccount', disccount))

    file_re = re.compile(".*\\.[mM][pP]3$")

    def understands(self, filename):
        return self.file_re.match(filename)

    def parse(self, filename):
        d = []
        mp3 = None
        try:
            mp3 = mutagen.mp3.MP3(filename)
        except mutagen.mp3.HeaderNotFoundError:
            pass
        except struct.error:
            pass
        if mp3 is not None:
            if mp3.tags is not None:
                self.handle_string_tags(self.mp3_string_map, mp3, d)
                self.handle_int_tags(self.mp3_int_map, mp3, d)
                self.handle_rating(mp3, d)
                self.handle_track(mp3, d)
                self.handle_disc(mp3, d)
            self.add_file_info(filename, d)
            d.extend([do('daap.songtime', mp3.info.length * 1000),
                      do('daap.songbitrate', mp3.info.bitrate / 1000),
                      do('daap.songsamplerate', mp3.info.sample_rate),
                      do('daap.songformat', 'mp3'),
                      do('daap.songdescription', 'MPEG Audio File'),
                      ])
            name = self.set_itemname_if_unset(os.path.basename(filename), d)
            return (d, name)
        else:
            return (None, None)



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import spydaap.parser.vorbis
import re
from spydaap.daap import do


class OggParser(spydaap.parser.vorbis.VorbisParser):
    file_re = re.compile(".*\\.[oO][gG][gG]$")

    def understands(self, filename):
        return self.file_re.match(filename)

    def parse_extra_vorbis(self, filename, md, daap):
        daap.extend([do('daap.songbitrate', md.info.bitrate / 1000),
                     do('daap.songformat', 'ogg'),
                     do('daap.songdescription', 'Ogg/Vorbis Audio File')])



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import os
import playlists
import spydaap.parser.mp3
import spydaap.parser.ogg
import spydaap.parser.flac

spydaap_dir = os.path.expanduser("~/.spydaap")
if not(os.path.isdir(spydaap_dir)):
    os.mkdir(spydaap_dir)

server_name = "spydaap"
port = 3689
media_path = os.path.expanduser("~/Music/")
cache_dir = os.path.join(spydaap_dir, "cache")
container_list = [playlists.Library()]
parsers = [spydaap.parser.mp3.Mp3Parser(),
           spydaap.parser.flac.FlacParser(),
           spydaap.parser.ogg.OggParser()]


class ContentRangeFile(object):

    def __init__(self, name, parent, start, end=None, chunk=1024):
        self.name = name
        self.parent = parent
        self.start = start
        self.end = end
        self.chunk = chunk
        self.parent.seek(self.start)
        self.read = start
        self.length = self.end - self.start

    def __len__(self):
        return self.length

    def next(self):
        to_read = self.chunk
        if (self.end is not None):
            if (self.read >= self.end):
                self.parent.close()
                raise StopIteration
            if (to_read + self.read > self.end):
                to_read = self.end - self.read
            retval = self.parent.read(to_read)
            self.read = self.read + len(retval)
        else:
            retval = self.parent.read(to_read)
        if retval == '':
            self.parent.close()
            raise StopIteration
        else:
            return retval

    def __iter__(self):
        return self



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import re
import spydaap
import os
from spydaap.daap import do


class AviParser(spydaap.parser.Parser):
    file_re = re.compile(".*\\.[aA][vV][iI]$")

    def understands(self, filename):
        return self.file_re.match(filename)

    def parse(self, filename):
        name = filename
        statinfo = os.stat(filename)
        d = [do('daap.songsize', statinfo.st_size),
             do('daap.songdateadded', statinfo.st_mtime),
             do('daap.songdatemodified', statinfo.st_mtime),
             do('dmap.itemname', name),
             do('daap.songalbum', ''),
             do('daap.songartist', ''),
             do('daap.songbitrate', 0),
             do('daap.songcomment', ''),
             do('daap.songdescription', 'MPEG-4 video file'),
             do('daap.songformat', 'avi'),
             do('com.apple.itunes.mediakind', 2),
             do('com.apple.itunes.has-video', True)
             ]
        return (d, name)



# daap.py
#
# DAAP classes and methods.
#
# original work (c) 2004, Davyd Madeley <davyd@ucc.asn.au>
#
# Later iTunes authentication work and object model
# copyright 2005 Tom Insam <tom@jerakeen.org>
#
# Stripped clean + a few bug fixes, Erik Hetzner

import logging
import struct
import sys

from spydaap.daap_data import (
    dmapDataTypes,
    dmapReverseDataTypes,
    dmapCodeTypes,
    dmapNames,
)

__all__ = ['DAAPError', 'DAAPObject', 'do']

log = logging.getLogger('daap')


def DAAPParseCodeTypes(treeroot):
    # the treeroot we are given should be a
    # dmap.contentcodesresponse
    if treeroot.codeName() != 'dmap.contentcodesresponse':
        raise DAAPError("DAAPParseCodeTypes: We cannot generate a dictionary from this tree.")
        return
    for object in treeroot.contains:
        # each item should be one of two things
        # a status code, or a dictionary
        if object.codeName() == 'dmap.status':
            pass
        elif object.codeName() == 'dmap.dictionary':
            code = None
            name = None
            dtype = None
            # a dictionary object should contain three items:
            # a 'dmap.contentcodesnumber' the 4 letter content code
            # a 'dmap.contentcodesname' the name of the code
            # a 'dmap.contentcodestype' the type of the code
            for info in object.contains:
                if info.codeName() == 'dmap.contentcodesnumber':
                    code = info.value
                elif info.codeName() == 'dmap.contentcodesname':
                    name = info.value
                elif info.codeName() == 'dmap.contentcodestype':
                    try:
                        dtype = dmapDataTypes[info.value]
                    except KeyError:
                        log.debug('DAAPParseCodeTypes: unknown data type %s for code %s, defaulting to s', info.value, name)
                        dtype = 's'
                else:
                    raise DAAPError('DAAPParseCodeTypes: unexpected code %s at level 2' % info.codeName())
            if code is None or name is None or dtype is None:
                log.debug('DAAPParseCodeTypes: missing information, not adding entry')
            else:
                try:
                    dtype = dmapReverseDataTypes[name]
                except KeyError:
                    pass
                dmapCodeTypes[code] = (name, dtype)
        else:
            raise DAAPError('DAAPParseCodeTypes: unexpected code %s at level 1' % info.codeName())


class DAAPError(Exception):
    pass


class DAAPObject(object):

    def __init__(self, code=None, value=None, **kwargs):
        if (code is not None):
            if (len(code) == 4):
                self.code = code
            else:
                self.code = dmapNames[code]
            if self.code is None or self.code not in dmapCodeTypes:
                self.type = None
            else:
                self.type = dmapCodeTypes[self.code][1]
            self.value = value
            if self.type == 'c' and isinstance(self.value, list):
                self.contains = value
        if 'parent' in kwargs:
            kwargs['parent'].contains.append(self)

    def getAtom(self, code):
        """returns an atom of the given code by searching 'contains' recursively."""
        if self.code == code:
            if self.type == 'c':
                return self
            return self.value

        # ok, it's not us. check our children
        if hasattr(self, 'contains'):
            for object in self.contains:
                value = object.getAtom(code)
                if value:
                    return value
        return None

    def codeName(self):
        if self.code is None or self.code not in dmapCodeTypes:
            return None
        else:
            return dmapCodeTypes[self.code][0]

    def objectType(self):
        if self.code is None or self.code not in dmapCodeTypes:
            return None
        else:
            return dmapCodeTypes[self.code][1]

    def printTree(self, level=0, out=sys.stdout):
        if hasattr(self, 'value'):
            out.write('\t' * level + '%s (%s)\t%s\t%s\n' % (self.codeName(), self.code, self.type, self.value))
        else:
            out.write('\t' * level + '%s (%s)\t%s\t%s\n' % (self.codeName(), self.code, self.type, None))
        if hasattr(self, 'contains'):
            for object in self.contains:
                object.printTree(level + 1)

    def encode(self):
        # generate DMAP tagged data format
        # step 1 - find out what type of object we are
        if self.type == 'c':
            # our object is a container,
            # this means we're going to have to
            # check contains[]
            value = ''
            for item in self.contains:
                # get the data stream from each of the sub elements
                if isinstance(item, str):
                    # preencoded
                    value += item
                else:
                    value += item.encode()
            # get the length of the data
            length = len(value)
            # pack: 4 byte code, 4 byte length, length bytes of value
            data = struct.pack('!4sI%ss' % length, self.code, length, value)
            return data
        else:
            # we don't have to traverse anything
            # to calculate the length and such
            # we want to encode the contents of
            # value for our value
            value = self.value
            if isinstance(value, float):
                value = int(value)
            if self.type == 'v':
                value = value.split('.')
                value = struct.pack('!HH', int(value[0]), int(value[1]))
                packing = "4s"
            elif self.type == 'l':
                packing = 'q'
            elif self.type == 'ul':
                packing = 'Q'
            elif self.type == 'i':
                if (isinstance(value, str) and len(value) <= 4):
                    packing = '4s'
                else:
                    packing = 'i'
            elif self.type == 'ui':
                packing = 'I'
            elif self.type == 'h':
                packing = 'h'
                if (value > 32767):
                    value = 32767
                elif (value < -32768):
                    value = -32768
            elif self.type == 'uh':
                packing = 'H'
            elif self.type == 'b':
                packing = 'b'
            elif self.type == 'ub':
                packing = 'B'
            elif self.type == 't':
                packing = 'I'
            elif self.type == 's':
                if isinstance(value, unicode):
                    value = value.encode('utf-8')
                packing = '%ss' % len(value)
            else:
                raise DAAPError('DAAPObject: encode: unknown code %s' % self.code)
                return
            # calculate the length of what we're packing
            length = struct.calcsize('!%s' % packing)
            # pack: 4 characters for the code, 4 bytes for the length, and 'length' bytes for the value
            data = struct.pack('!4sI%s' % (packing), self.code, length, value)
            return data

    def processData(self, str):
        # read 4 bytes for the code and 4 bytes for the length of the objects data
        data = str.read(8)

        if not data:
            return
        self.code, self.length = struct.unpack('!4sI', data)

        # now we need to find out what type of object it is
        if self.code is None or self.code not in dmapCodeTypes:
            self.type = None
        else:
            self.type = dmapCodeTypes[self.code][1]

        if self.type == 'c':
            start_pos = str.tell()
            self.contains = []
            # the object is a container, we need to pass it
            # it's length amount of data for processessing
            while str.tell() < start_pos + self.length:
                object = DAAPObject()
                self.contains.append(object)
                object.processData(str)
            return

        # not a container, we're a single atom. Read it.
        code = str.read(self.length)

        if self.type == 'l':
            # the object is a long long number,
            self.value = struct.unpack('!q', code)[0]
        elif self.type == 'ul':
            # the object is an unsigned long long
            self.value = struct.unpack('!Q', code)[0]
        elif self.type == 'i':
            # the object is a number,
            self.value = struct.unpack('!i', code)[0]
        elif self.type == 'ui':
            # unsigned integer
            self.value = struct.unpack('!I', code)[0]
        elif self.type == 'h':
            # this is a short number,
            self.value = struct.unpack('!h', code)[0]
        elif self.type == 'uh':
            # unsigned short
            self.value = struct.unpack('!H', code)[0]
        elif self.type == 'b':
            # this is a byte long number
            self.value = struct.unpack('!b', code)[0]
        elif self.type == 'ub':
            # unsigned byte
            self.value = struct.unpack('!B', code)[0]
        elif self.type == 'v':
            # this is a version tag
            self.value = float("%s.%s" % struct.unpack('!HH', code))
        elif self.type == 't':
            # this is a time string
            self.value = struct.unpack('!I', code)[0]
        elif self.type == 's':
            # the object is a string
            # we need to read length characters from the string
            try:
                self.value = unicode(
                    struct.unpack('!%ss' % self.length, code)[0], 'utf-8')
            except UnicodeDecodeError:
                # oh, urgh
                self.value = unicode(
                    struct.unpack('!%ss' % self.length, code)[0], 'latin-1')
        else:
            # we don't know what to do with this object
            # put it's raw data into value
            log.debug('DAAPObject: Unknown code %s for type %s, writing raw data', code, self.code)
            self.value = code

do = DAAPObject



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

from hashlib import md5
import os


class Cache(object):

    def __init__(self, dir):
        self.dir = os.path.abspath(dir)
        if (not(os.path.exists(self.dir))):
            os.mkdir(self.dir)

    def get(self, id, func):
        id = md5(id).hexdigest()
        fn = os.path.join(self.dir, id)
        if (not(os.path.exists(fn))):
            f = open(fn, 'w')
            func(f)
            f.close()
        return open(fn)

    def clean(self):
        for f in os.listdir(self.dir):
            p = os.path.join(self.dir, f)
            if os.path.isfile(p):
                os.remove(p)


class OrderedCache(object):

    class Iter:

        def __init__(self, cache):
            self.cache = cache
            self.n = 0

        def __iter__(self):
            return self

        def next(self):
            if self.n >= len(self.cache):
                raise StopIteration
            self.n = self.n + 1
            return self.cache.get_item_by_id(self.n)

    def __init__(self, dir):
        self.dir = os.path.abspath(dir)
        if (not(os.path.exists(self.dir))):
            os.mkdir(self.dir)

    def __iter__(self):
        return OrderedCache.Iter(self)

    def get_item_by_id(self, id):
        fi = open(os.path.join(self.dir, 'index'), 'r')
        fi.seek((int(id) - 1) * 32)
        cfn = fi.read(32)
        fi.close()
        return self.get_item_by_pid(cfn, id)

    def __len__(self):
        return os.path.getsize(os.path.join(self.dir, 'index')) / 32

    def build_index(self, pid_list=None):
        index_fn = os.path.join(self.dir, 'index')
        if os.path.exists(index_fn):
            os.remove(index_fn)
        if pid_list is None:
            pid_list = sorted([f for f in os.listdir(self.dir) if f != "index"])
        fi = open(index_fn, 'w')
        for pid in pid_list:
            fi.write(pid)
        fi.close()

    def clean(self):
        for f in os.listdir(self.dir):
            p = os.path.join(self.dir, f)
            if os.path.isfile(p):
                os.remove(p)


class OrderedCacheItem(object):

    def __init__(self, cache, pid, id):
        self.cache = cache
        self.pid = pid
        self.id = id
        self.path = os.path.join(self.cache.dir, pid)

    def get_mtime(self):
        return os.stat(self.path).st_mtime

    def get_exists(self):
        return os.path.exists(self.path)

    def get_id(self):
        return self.id

    def get_pid(self):
        return self.pid



#!/usr/bin/env python
# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import optparse

import BaseHTTPServer
import SocketServer
import grp
import os
import pwd
import select
import signal
import spydaap
import sys
import socket
# import httplib, logging
import spydaap.daap
import spydaap.metadata
import spydaap.containers
import spydaap.cache
import spydaap.server
import spydaap.zeroconf

config_file = os.path.join(spydaap.spydaap_dir, "config.py")
if os.path.isfile(config_file):
    execfile(config_file)

cache = spydaap.cache.Cache(spydaap.cache_dir)
md_cache = spydaap.metadata.MetadataCache(os.path.join(spydaap.cache_dir, "media"), spydaap.parsers)
container_cache = spydaap.containers.ContainerCache(os.path.join(spydaap.cache_dir, "containers"), spydaap.container_list)
keep_running = True


class Log(object):
    """file like for writes with auto flush after each write
    to ensure that everything is logged, even during an
    unexpected exit."""

    def __init__(self, f, quiet):
        self.f = f
        self.quiet = quiet
        self.stdout = sys.__stdout__

    def write(self, s):
        self.f.write(s)
        self.f.flush()
        if not self.quiet:
            self.stdout.write(s)
            self.stdout.flush()


class MyThreadedHTTPServer(SocketServer.ThreadingMixIn, BaseHTTPServer.HTTPServer):
    """Handle requests in a separate thread."""
    timeout = 1

    def __init__(self, *args):
        BaseHTTPServer.HTTPServer.__init__(self, *args)
        self.keep_running = True

    def serve_forever(self):
        while self.keep_running:
            self.handle_request()

    def force_stop(self):
        self.keep_running = False
        self.server_close()


def rebuild_cache(signum=None, frame=None):
    md_cache.build(os.path.abspath(spydaap.media_path))
    container_cache.clean()
    container_cache.build(md_cache)
    cache.clean()


def make_shutdown(httpd):
    def _shutdown(signum, frame):
        httpd.force_stop()
    return _shutdown

# invalid default pid ; prevents killing something unintended


def really_main(opts, parent_pid=99999999999999):
    rebuild_cache()
    zeroconf = spydaap.zeroconf.Zeroconf(spydaap.server_name,
                                         spydaap.port,
                                         stype="_daap._tcp")
    zeroconf.publish()
    try:
        httpd = MyThreadedHTTPServer(('0.0.0.0', spydaap.port),
                                     spydaap.server.makeDAAPHandlerClass(spydaap.server_name, cache, md_cache, container_cache))
        # write pid to pidfile
        open(opts.pidfile, 'w').write("%d" % parent_pid)
    except socket.error:
        if not opts.daemonize:
            print "Another DAAP server is already running. Exiting."

        sys.exit(0)  # silently exit; another instance is already running

    signal.signal(signal.SIGTERM, make_shutdown(httpd))
    signal.signal(signal.SIGHUP, rebuild_cache)

    while httpd.keep_running:
        try:
            httpd.serve_forever()
        except select.error:
            pass
        except KeyboardInterrupt:
            httpd.force_stop()
    zeroconf.unpublish()


def main():
    def getpwname(o, s, value, parser):
        parser.values.user = pwd.getpwnam(value)[2]

    def getgrname(o, s, value, parser):
        parser.values.group = grp.getgrnam(value)[2]

    parser = optparse.OptionParser()

    parser.add_option("-d", "--daemon", action="store_true",
                      dest="daemonize", default=False,
                      help="run in the background as a daemon process")

    parser.add_option("-k", "--kill", action="store_true",
                      dest="kill_daemon", default=False,
                      help="kill a running daemon process")

    parser.add_option("-n", "--servername", dest="servername",
                      default=None,
                      help="set the server-name (must be < 64 chars); default is 'spydaap'")

    parser.add_option("-f", "--folder", dest="folderpath",
                      default=None,
                      help="set the path to the media folder (default is ~/Music)")

    parser.add_option("-q", "--quiet", action="store_true",
                      dest="quiet", default=False,
                      help="suppress logging to stdout")

    parser.add_option("-g", "--group", dest="group", action="callback",
                      help="specify group to run as", type="str",
                      callback=getgrname, default=os.getgid())

    parser.add_option("-u", "--user", dest="user", action="callback",
                      help="specify username to run as", type="string",
                      callback=getpwname, default=os.getuid())

    parser.add_option("-l", "--logfile", dest="logfile",
                      default=os.path.join(spydaap.spydaap_dir, "spydaap.log"),
                      help="use log file (default is ~/.spydaap/spydaap.log)")

    parser.add_option("-p", "--pidfile", dest="pidfile",
                      default=os.path.join(spydaap.spydaap_dir, "spydaap.pid"),
                      help="use pid file (default is ~/.spydaap/spydaap.pid)")

    opts, args = parser.parse_args()

    if opts.user == 0 or opts.group == 0:
        sys.stderr.write("spydaap must not run as root\n")
        sys.exit(2)
    # ensure the that the daemon runs a normal user
    os.setegid(opts.group)
    os.seteuid(opts.user)

    if opts.kill_daemon:
        try:
            pid = int(open(opts.pidfile, 'r').read())
            os.kill(pid, signal.SIGTERM)
            print "Daemon killed."
        except (OSError, IOError):
            print "Unable to kill daemon -- not running, or missing pid file?"

        sys.exit(0)

    if opts.servername is not None:
        spydaap.server_name = opts.servername

    if len(spydaap.server_name) > 63:
        # truncate to max valid length (63 characters)
        spydaap.server_name = spydaap.server_name[:63]

    if opts.folderpath is not None:
        spydaap.media_path = os.path.expanduser(opts.folderpath)

    if not(opts.daemonize):
        if not opts.quiet:
            print "spydaap server started (use --help for more options).  Press Ctrl-C to exit."
        # redirect outputs to a logfile
        sys.stdout = sys.stderr = Log(open(opts.logfile, 'a+'), opts.quiet)
        really_main(opts)
    else:
        if not opts.quiet:
            print "spydaap daemon started in background."
        # redirect outputs to a logfile
        sys.stdout = sys.stderr = Log(open(opts.logfile, 'a+'), True)
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # decouple from parent environment
        os.chdir("/")  # don't prevent unmounting....
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            # if in second parent, exit from it
            if pid > 0:
                # store pid temporarily (don't overwrite real pidfile until we
                # know server has successfully started)
                open(opts.pidfile + '.tmp', 'w').write("%d" % pid)
                parent_pid = pid
                sys.exit(0)
        except OSError as e:
            print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror)
            sys.exit(1)
        # load parent pid
        parent_pid = int(open(opts.pidfile + '.tmp', 'r').read())

        really_main(opts, parent_pid)

if __name__ == "__main__":
    main()



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import os
import time


class Playlist(object):
    name = None
    smart_playlist = True

    def sort(self, entries):
        pass

    def safe_cmp(self, a, b, key):
        if key in a and key in b:
            return cmp(a[key], b[key])
        elif key in a:
            return 1
        elif key in b:
            return -1
        else:
            return 0

    def safe_cmp_series(self, a, b, key_list):
        if len(key_list) == 0:
            return 0
        key = key_list[0]
        r = self.safe_cmp(a, b, key)
        if r != 0:
            return r
        else:
            return self.safe_cmp_series(a, b, key_list[1:])


class Library(Playlist):

    def __init__(self):
        self.name = "Library"
        self.smart_playlist = False

    def contains(self, md):
        return True


class Genre(Playlist):

    def __init__(self, name, genre):
        self.name = name
        self.genre = genre

    def contains(self, md):
        if not('daap.songgenre' in md):
            return False
        else:
            songgenre = md['daap.songgenre'].lower()
            if isinstance(self.genre, str):
                return songgenre == self.genre
            elif isinstance(self.genre, list):
                return songgenre in self.genre


class YearRange(Playlist):

    def __init__(self, name, first, last=None):
        self.name = name
        if last is None:
            last = first
        self.last = last
        self.first = first

    def contains(self, md):
        if not('daap.songyear' in md):
            return False
        else:
            year = md['daap.songyear']
            return year >= self.first and year <= self.last

    def sort(self, entries):
        def s(a, b):
            return self.safe_cmp_series(a, b, ['daap.songyear',
                                               'daap.songartist',
                                               'daap.songalbum',
                                               'daap.songtracknumber'])
        entries.sort(cmp=s)


class Recent(Playlist):

    def __init__(self, name, seconds=604800):
        self.name = name
        self.seconds = seconds

    def contains(self, md):
        f_mtime = os.stat(md.get_original_filename()).st_mtime
        return ((f_mtime + self.seconds) > time.time())


class Rating(Playlist):

    def __init__(self, name, rating):
        self.name = name
        self.rating = rating

    def contains(self, md):
        if 'daap.songuserrating' in md:
            return md['daap.songuserrating'] >= self.rating
        else:
            return False


class FilenameContains(Playlist):

    def __init__(self, name):
        self.name = name

    def contains(self, md):
        return self.name in md.get_original_filename()




dmapCodeTypes = {
    #'f\x8dch': ('dmap.haschildcontainers', 'b'),
    'abal': ('daap.browsealbumlisting', 'c'),
    'abar': ('daap.browseartistlisting', 'c'),
    'abcp': ('daap.browsecomposerlisting', 'c'),
    'abgn': ('daap.browsegenrelisting', 'c'),
    'abpl': ('daap.baseplaylist', 'b'),
    'abro': ('daap.databasebrowse', 'c'),
    'adbs': ('daap.databasesongs', 'c'),
    'aeAI': ('com.apple.itunes.itms-artistid', 'i'),
    'aeCI': ('com.apple.itunes.itms-composerid', 'i'),
    'aeEN': ('com.apple.itunes.episode-num-str', 's'),
    'aeES': ('com.apple.itunes.episode-sort', 'i'),
    'aeFP': ('com.apple.itunes.req-fplay', 'b'),
    'aeGU': ('com.apple.itunes.gapless-dur', 'l'),
    'aeGD': ('com.apple.itunes.gapless-enc-dr', 'i'),
    'aeGE': ('com.apple.itunes.gapless-enc-del', 'i'),
    'aeGH': ('com.apple.itunes.gapless-heur', 'i'),
    'aeGI': ('com.apple.itunes.itms-genreid', 'i'),
    'aeGR': ('com.apple.itunes.gapless-resy', 'l'),
    'aeHV': ('com.apple.itunes.has-video', 'b'),
    'aeMK': ('com.apple.itunes.mediakind', 'b'),
    'aeNN': ('com.apple.itunes.network-name', 's'),
    'aeNV': ('com.apple.itunes.norm-volume', 'i'),
    'aePC': ('com.apple.itunes.is-podcast', 'b'),
    'aePI': ('com.apple.itunes.itms-playlistid', 'i'),
    'aePP': ('com.apple.itunes.is-podcast-playlist', 'b'),
    'aePS': ('com.apple.itunes.special-playlist', 'b'),
    'aeSU': ('com.apple.itunes.season-num', 'i'),
    'aeSF': ('com.apple.itunes.itms-storefrontid', 'i'),
    'aeSI': ('com.apple.itunes.itms-songid', 'i'),
    'aeSN': ('com.apple.itunes.series-name', 's'),
    'aeSP': ('com.apple.itunes.smart-playlist', 'b'),
    'aeSV': ('com.apple.itunes.music-sharing-version', 'i'),
    'agrp': ('daap.songgrouping', 's'),
    'aply': ('daap.databaseplaylists', 'c'),
    'aprm': ('daap.playlistrepeatmode', 'b'),
    'apro': ('daap.protocolversion', 'v'),
    'apsm': ('daap.playlistshufflemode', 'b'),
    'apso': ('daap.playlistsongs', 'c'),
    'arif': ('daap.resolveinfo', 'c'),
    'arsv': ('daap.resolve', 'c'),
    'asaa': ('daap.songalbumartist', 's'),
    'asal': ('daap.songalbum', 's'),
    'asar': ('daap.songartist', 's'),
    'asbk': ('daap.bookmarkable', 'b'),
    'asbo': ('daap.songbookmark', 'i'),
    'asbr': ('daap.songbitrate', 'h'),
    'asbt': ('daap.songbeatsperminute', 'h'),
    'ascd': ('daap.songcodectype', 'i'),
    'ascm': ('daap.songcomment', 's'),
    'ascn': ('daap.songcontentdescription', 's'),
    'asco': ('daap.songcompilation', 'b'),
    'ascp': ('daap.songcomposer', 's'),
    'ascr': ('daap.songcontentrating', 'b'),
    'ascs': ('daap.songcodecsubtype', 'i'),
    'asct': ('daap.songcategory', 's'),
    'asda': ('daap.songdateadded', 't'),
    'asdb': ('daap.songdisabled', 'b'),
    'asdc': ('daap.songdisccount', 'h'),
    'asdk': ('daap.songdatakind', 'b'),
    'asdm': ('daap.songdatemodified', 't'),
    'asdn': ('daap.songdiscnumber', 'h'),
    'asdp': ('daap.songdatepurchased', 't'),
    'asdr': ('daap.songdatereleased', 't'),
    'asdt': ('daap.songdescription', 's'),
    'ased': ('daap.songextradata', 'h'),
    'aseq': ('daap.songeqpreset', 's'),
    'asfm': ('daap.songformat', 's'),
    'asgn': ('daap.songgenre', 's'),
    'asgp': ('daap.songgapless', 'b'),
    'ashp': ('daap.songhasbeenplayed', 'b'),
    'asky': ('daap.songkeywords', 's'),
    'aslc': ('daap.songlongcontentdescription', 's'),
    'asrv': ('daap.songrelativevolume', 'ub'),
    'assu': ('daap.sortalbum', 's'),
    'assa': ('daap.sortartist', 's'),
    'assc': ('daap.sortcomposer', 's'),
    'assl': ('daap.sortalbumartist', 's'),
    'assn': ('daap.sortname', 's'),
    'assp': ('daap.songstoptime', 'i'),
    'assr': ('daap.songsamplerate', 'i'),
    'asss': ('daap.sortseriesname', 's'),
    'asst': ('daap.songstarttime', 'i'),
    'assz': ('daap.songsize', 'i'),
    'astc': ('daap.songtrackcount', 'h'),
    'astm': ('daap.songtime', 'i'),
    'astn': ('daap.songtracknumber', 'h'),
    'asul': ('daap.songdataurl', 's'),
    'asur': ('daap.songuserrating', 'b'),
    'asyr': ('daap.songyear', 'h'),
    'ated': ('daap.supportsextradata', 'h'),
    'avdb': ('daap.serverdatabases', 'c'),
    'mbcl': ('dmap.bag', 'c'),
    'mccr': ('dmap.contentcodesresponse', 'c'),
    'mcna': ('dmap.contentcodesname', 's'),
    'mcnm': ('dmap.contentcodesnumber', 'i'),
    'mcon': ('dmap.container', 'c'),
    'mctc': ('dmap.containercount', 'i'),
    'mcti': ('dmap.containeritemid', 'i'),
    'mcty': ('dmap.contentcodestype', 'h'),
    'mdcl': ('dmap.dictionary', 'c'),
    'miid': ('dmap.itemid', 'i'),
    'mikd': ('dmap.itemkind', 'b'),
    'mimc': ('dmap.itemcount', 'i'),
    'minm': ('dmap.itemname', 's'),
    'mlcl': ('dmap.listing', 'c'),
    'mlid': ('dmap.sessionid', 'i'),
    'mlit': ('dmap.listingitem', 'c'),
    'mlog': ('dmap.loginresponse', 'c'),
    'mpco': ('dmap.parentcontainerid', 'i'),
    'mper': ('dmap.persistentid', 'l'),
    'mpro': ('dmap.protocolversion', 'v'),
    'mrco': ('dmap.returnedcount', 'i'),
    'msau': ('dmap.authenticationmethod', 'b'),
    'msal': ('dmap.supportsautologout', 'b'),
    'msas': ('dmap.authenticationschemes', 'i'),
    'msbr': ('dmap.supportsbrowse', 'b'),
    'msdc': ('dmap.databasescount', 'i'),
    'msex': ('dmap.supportsextensions', 'b'),
    'msix': ('dmap.supportsindex', 'b'),
    'mslr': ('dmap.loginrequired', 'b'),
    'mspi': ('dmap.supportspersistentids', 'b'),
    'msqy': ('dmap.supportsquery', 'b'),
    'msrs': ('dmap.supportsresolve', 'b'),
    'msrv': ('dmap.serverinforesponse', 'c'),
    'mstc': ('dmap.utctime', 't'),
    'mstm': ('dmap.timeoutinterval', 'i'),
    'msto': ('dmap.utcoffset', 'ui'),
    'msts': ('dmap.statusstring', 's'),
    'mstt': ('dmap.status', 'i'),
    'msup': ('dmap.supportsupdate', 'b'),
    'mtco': ('dmap.specifiedtotalcount', 'i'),
    'mudl': ('dmap.deletedidlisting', 'c'),
    'mupd': ('dmap.updateresponse', 'c'),
    'musr': ('dmap.serverrevision', 'i'),
    'muty': ('dmap.updatetype', 'b'),
    'ppro': ('dpap.protocolversion', 'i'),
    'pret': ('dpap.unknown', 'c'),
}

dmapDataTypes = {
    1: 'b',  # byte
    2: 'ub',  # unsigned byte
    3: 'h',  # short
    4: 'uh',  # unsigned short
    5: 'i',  # integer
    6: 'ui',  # unsigned integer
    7: 'l',  # long
    8: 'ul',  # unsigned long
    9: 's',  # string
    10: 't',  # timestamp
    11: 'v',  # version
    12: 'c',  # container
}

dmapNames = {}
for k in dmapCodeTypes.keys():
    dmapNames[dmapCodeTypes[k][0]] = k

dmapReverseDataTypes = {}
for k in dmapDataTypes.keys():
    dmapReverseDataTypes[dmapDataTypes[k]] = k



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

from hashlib import md5

import os
import struct
import spydaap.cache
import StringIO
from spydaap.daap import do


class MetadataCache(spydaap.cache.OrderedCache):

    def __init__(self, cache_dir, parsers):
        self.parsers = parsers
        super(MetadataCache, self).__init__(cache_dir)

    def get_item_by_pid(self, pid, n=None):
        return MetadataCacheItem(self, pid, n)

    def build(self, dir, marked={}, link=False):
        for path, dirs, files in os.walk(dir):
            for d in dirs:
                if os.path.islink(os.path.join(path, d)):
                    self.build(os.path.join(path, d), marked, True)
            for fn in files:
                ffn = os.path.join(path, fn)
                digest = md5(ffn).hexdigest()
                marked[digest] = True
                md = self.get_item_by_pid(digest)
                if (not(md.get_exists()) or
                        (md.get_mtime() < os.stat(ffn).st_mtime)):
                    for p in self.parsers:
                        if p.understands(ffn):
                            try:
                                (m, name) = p.parse(ffn)
                                if m is not None:
                                    MetadataCacheItem.write_entry(self.dir,
                                                                  name, ffn, m)
                            except:
                                pass
        if not(link):
            for item in os.listdir(self.dir):
                if (len(item) == 32) and not(item in marked):
                    os.remove(os.path.join(self.dir, item))
            self.build_index()


class MetadataCacheItem(spydaap.cache.OrderedCacheItem):

    @classmethod
    def write_entry(self, dir, name, fn, daap):
        if isinstance(name, unicode):
            name = name.encode('utf-8')
        data = "".join([d.encode() for d in daap])
        data = struct.pack('!i%ss' % len(name), len(name), name) + data
        data = struct.pack('!i%ss' % len(fn), len(fn), fn) + data
        cachefn = os.path.join(dir, md5(fn).hexdigest())
        f = open(cachefn, 'w')
        f.write(data)
        f.close()

    def __init__(self, cache, pid, id):
        super(MetadataCacheItem, self).__init__(cache, pid, id)
        self.file = None
        self.daap = None
        self.name = None
        self.original_filename = None
        self.daap_raw = None
        self.md = None

    def __getitem__(self, k):
        return self.md[k]

    def has_key(self, k):
        return k in self.get_md()

    def read(self):
        f = open(self.path)
        fn_len = struct.unpack('!i', f.read(4))[0]
        self.original_filename = f.read(fn_len)
        name_len = struct.unpack('!i', f.read(4))[0]
        self.name = f.read(name_len)
        self.daap_raw = f.read()
        f.close()

    def get_original_filename(self):
        if self.original_filename is None:
            self.read()
        return self.original_filename

    def get_name(self):
        if self.name is None:
            self.read()
        return self.name

    def get_dmap_raw(self):
        if self.daap_raw is None:
            self.read()
        return self.daap_raw

    def get_md(self):
        if self.md is None:
            self.md = {}
            s = StringIO.StringIO(self.get_dmap_raw())
            l = len(self.get_dmap_raw())
            data = []
            while s.tell() != l:
                d = do()
                d.processData(s)
                data.append(d)
            for d in data:
                self.md[d.codeName()] = d.value
        return self.md



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

from hashlib import md5
import os
import struct
import spydaap.cache
from spydaap.daap import do


class ContainerCache(spydaap.cache.OrderedCache):

    def __init__(self, cache_dir, container_list):
        self.container_list = container_list
        super(ContainerCache, self).__init__(cache_dir)

    def get_item_by_pid(self, pid, n=None):
        return ContainerCacheItem(self, pid, None)

    def build(self, md_cache):
        def build_do(md, id):
            d = do('dmap.listingitem',
                   [do('dmap.itemkind', 2),
                    do('dmap.itemid', md.id),
                    do('dmap.itemname', md.get_name()),
                    do('dmap.containeritemid', id),
                    do('daap.songformat', md.get_md()['daap.songformat'])])
            return d
        pid_list = []
        for pl in self.container_list:
            entries = [n for n in md_cache if pl.contains(n)]
            pl.sort(entries)
            d = do('daap.playlistsongs',
                   [do('dmap.status', 200),
                    do('dmap.updatetype', 0),
                    do('dmap.specifiedtotalcount', len(entries)),
                    do('dmap.returnedcount', len(entries)),
                    do('dmap.listing',
                        [build_do(md, id) for (id, md) in enumerate(entries)])
                    ])
            ContainerCacheItem.write_entry(self.dir, pl.name, d, len(entries))
            pid_list.append(md5(pl.name).hexdigest())
        self.build_index(pid_list)


class ContainerCacheItem(spydaap.cache.OrderedCacheItem):

    @classmethod
    def write_entry(self, dir, name, d, length):
        data = struct.pack('!i', length)
        data = data + struct.pack('!i%ss' % len(name), len(name), name)
        data = data + d.encode()
        cachefn = os.path.join(dir, md5(name).hexdigest())
        f = open(cachefn, 'w')
        f.write(data)
        f.close()

    def __init__(self, cache, pid, id):
        super(ContainerCacheItem, self).__init__(cache, pid, id)
        self.daap_raw = None
        self.name = None
        self._len = None

    def read(self):
        f = open(self.path)
        self._len = struct.unpack('!i', f.read(4))[0]
        name_len = struct.unpack('!i', f.read(4))[0]
        self.name = f.read(name_len)
        self.daap_raw = f.read()
        f.close()

    def get_daap_raw(self):
        if self.daap_raw is None:
            self.read()
        return self.daap_raw

    def get_name(self):
        if self.name is None:
            self.read()
        return self.name

    def __len__(self):
        if self._len is None:
            self.read()
        return self._len



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import os
import re
from spydaap.daap import do


class Parser:

    def handle_string_tags(self, map, md, daap):
        h = {}
        for k in md.tags.keys():
            if k in map:
                tag = [unicode(t) for t in md.tags[k]]
                tag = [t for t in tag if t != ""]
                if not(map[k] in h):
                    h[map[k]] = []
                h[map[k]] = h[map[k]] + tag
        for k in h.keys():
            h[k].sort()
            daap.append(do(k, "/".join(h[k])))

    def handle_int_tags(self, map, md, daap):
        for k in md.tags.keys():
            if k in map:
                val = md.tags[k]
                if isinstance(val, list):
                    val = val[0]
                intval = self.my_int(unicode(val))
                if intval:
                    daap.append(do(map[k], intval))

    def add_file_info(self, filename, daap):
        statinfo = os.stat(filename)
        daap.extend([do('daap.songsize', os.path.getsize(filename)),
                     do('daap.songdateadded', statinfo.st_ctime),
                     do('daap.songdatemodified', statinfo.st_ctime)])

    def set_itemname_if_unset(self, name, daap):
        for d in daap:
            if d.code == 'minm':
                return d.value
        daap.extend([do('minm', name)])
        return name

    def clean_int_string(self, s):
        return re.sub(u'[^0-9]', '', unicode(s))

    def my_int(self, s):
        try:
            return int(self.clean_int_string(s))
        except Exception:
            return None



# Copyright (C) 2008 Erik Hetzner

# This file is part of Spydaap. Spydaap is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# Spydaap is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Spydaap. If not, see <http://www.gnu.org/licenses/>.

import mutagen
import spydaap
import re
import os
from spydaap.daap import do

# * TODO Implement song.songtrackcount, song.disccount
# daap.songbeatsperminute
# daap.songcomment
# daap.songdateadded
# daap.songdatemodified,
# daap.songdisabled,
# daap.songeqpreset
# daap.songformat
# daap.songdescription
# daap.songrelativevolume,
# daap.songsize,
# daap.songstarttime,
# daap.songstoptime,
# daap.songtime,
# daap.songuserrating,
# daap.songdatakind,
# daap.songdataurl


class VorbisParser(spydaap.parser.Parser):
    vorbis_string_map = {
        'grouping': 'daap.songgrouping',
        'title': 'dmap.itemname',
        'artist': 'daap.songartist',
        'composer': 'daap.songcomposer',
        'genre': 'daap.songgenre',
        'album': 'daap.songalbum',
        'albumartist': 'daap.songalbumartist',
    }

    vorbis_int_map = {
        'bpm': 'daap.songbeatsperminute',
        'date': 'daap.songyear',
        'year': 'daap.songyear',
        'compilation': 'daap.songcompilation',
    }

    def handle_track(self, flac, d):
        tracknumber = None
        trackcount = None
        if 'tracknumber' in flac.tags:
            t = str(flac.tags['tracknumber']).split('/')
            tracknumber = self.my_int(t[0])
            if (len(t) == 2):
                trackcount = self.my_int(t[1])
        if 'tracktotal' in flac.tags:
            trackcount = self.my_int(flac.tags['tracktotal'])
        if tracknumber:
            d.append(do('daap.songtracknumber', tracknumber))
        if trackcount:
            d.append(do('daap.songtrackcount', trackcount))

    def handle_disc(self, flac, d):
        discnumber = None
        disccount = None
        if 'discnumber' in flac.tags:
            t = unicode(flac.tags['discnumber'][0]).split('/')
            discnumber = self.my_int(t[0])
            if (len(t) == 2):
                disccount = self.my_int(t[1])
        if 'disctotal' in flac.tags:
            disccount = self.my_int(flac.tags['disctotal'])
        if discnumber:
            d.append(do('daap.songdiscnumber', discnumber))
        if disccount:
            d.append(do('daap.songdisccount', disccount))

    file_re = re.compile(".*\\.([fF][lL][aA][cC]|[oO][gG]{2})$")

    def understands(self, filename):
        return self.file_re.match(filename)

    def parse(self, filename):
        md = mutagen.File(filename)
        d = []
        if md.tags is not None:
            self.handle_string_tags(self.vorbis_string_map, md, d)
            self.handle_int_tags(self.vorbis_int_map, md, d)
            self.handle_track(md, d)
            self.handle_disc(md, d)
        self.add_file_info(filename, d)
        d.extend([do('daap.songtime', md.info.length * 1000),
                  do('daap.songsamplerate', md.info.sample_rate)])
        name = self.set_itemname_if_unset(os.path.basename(filename), d)
        if hasattr(self, 'parse_extra_vorbis'):
            self.parse_extra_vorbis(filename, md, d)
        return (d, name)



from setuptools import setup
import platform

reqs = ["mutagen>=1.2"]
# use avahi if available
try:
    import avahi
except ImportError:
    reqs.append("pybonjour>=1.1")

setup(
    name="spydaap",
    version="0.1",
    author="Erik Hetzner",
    author_email="egh@e6h.org",
    description="A simple DAAP server",
    long_description="""
=========
 spydaap
=========

Spydaap is a media server supporting the DAAP protocol (aka iTunes
sharing). It is written in Python, uses the mutagen media metadata
library, and either the Avahi or pybonjour Zeroconf implementation.

Features:

 - Runs on Unix-like systems (Linux, *BSD, Mac OS X).
 - Can stream mp3s, ogg, flac, and Quicktime videos.
 - Supports "smart" playlists written in Python.
 - Written in 100 percent Python and easily modifiable.
 - Caches almost everything for fast performance.
 - Embeddable.

Note: This pypi package is maintained by the Exaile project, but will be kept in
sync with the upstream spydaap project. 
 
""",
    url="https://github.com/egh/spydaap",
    install_requires=reqs,
    entry_points={
        'console_scripts': [
            'spydaap=spydaap.cli:main'
        ]
    },
    packages=["spydaap", "spydaap.parser"],
    classifiers=["Development Status :: 4 - Beta",
                 "License :: OSI Approved :: GNU General Public License (GPL)",
                 "Programming Language :: Python",
                 "Topic :: Multimedia :: Sound/Audio",
                 "Operating System :: POSIX",
                 "Intended Audience :: End Users/Desktop"])



# adopted from http://stackp.online.fr/?p=35

__all__ = ["Zeroconf"]

import select
import logging

logger = logging.getLogger(__name__)


class Zeroconf(object):
    """A simple class to publish a network service with zeroconf using
    avahi or pybonjour, preferring pybonjour.
    """

    class Helper(object):

        def __init__(self, name, port, **kwargs):
            self.name = name
            self.port = port
            self.stype = kwargs.get('stype', "_http._tcp")
            self.domain = kwargs.get('domain', '')
            self.host = kwargs.get('host', '')
            self.text = kwargs.get('text', '')

    class Pybonjour(Helper):

        def publish(self):
            import pybonjour
            # records as in mt-daapd
            txtRecord = pybonjour.TXTRecord()
            txtRecord['txtvers'] = '1'
            txtRecord['iTSh Version'] = '131073'  # '196609'
            txtRecord['Machine Name'] = self.name
            txtRecord['Password'] = '0'  # 'False' ?
            # txtRecord['Database ID']        = '' # 16 hex digits
            #txtRecord['Version']            = '196616'
            # txtRecord['iTSh Version']       =
            # txtRecord['Machine ID']         = '' # 12 hex digits
            #txtRecord['Media Kinds Shared'] = '0'
            # txtRecord['OSsi']               = '0x1F6' #?
            # txtRecord['MID']                = '0x3AA6175DD7155BA7', = database id - 2 ?
            #txtRecord['dmv']                = '131077'

            def register_callback(sdRef, flags, errorCode, name, regtype, domain):
                pass

            self.sdRef = pybonjour.DNSServiceRegister(name=self.name,
                                                      regtype="_daap._tcp",
                                                      port=self.port,
                                                      callBack=register_callback,
                                                      txtRecord=txtRecord)

            while True:
                ready = select.select([self.sdRef], [], [])
                if self.sdRef in ready[0]:
                    pybonjour.DNSServiceProcessResult(self.sdRef)
                    break

        def unpublish(self):
            self.sdRef.close()

    class Avahi(Helper):

        def publish(self, ipv4=True, ipv6=True):
            import dbus
            import avahi
            bus = dbus.SystemBus()
            server = dbus.Interface(
                bus.get_object(
                    avahi.DBUS_NAME,
                    avahi.DBUS_PATH_SERVER),
                avahi.DBUS_INTERFACE_SERVER)

            self.group = dbus.Interface(
                bus.get_object(avahi.DBUS_NAME,
                               server.EntryGroupNew()),
                avahi.DBUS_INTERFACE_ENTRY_GROUP)

            if ipv4 and ipv6:
                proto = avahi.PROTO_UNSPEC
            elif ipv6:
                proto = avahi.PROTO_INET6
            else:  # we don't let them both be false
                proto = avahi.PROTO_INET

            self.group.AddService(avahi.IF_UNSPEC, proto,
                                  dbus.UInt32(0), self.name, self.stype, self.domain,
                                  self.host, dbus.UInt16(self.port), self.text)
            self.group.Commit()

        def unpublish(self):
            self.group.Reset()

    def __init__(self, *args, **kwargs):
        try:
            import pybonjour
            self.helper = Zeroconf.Pybonjour(*args, **kwargs)
        except ImportError:
            logger.info('pybonjour not found, using avahi')
            try:
                import avahi
                import dbus
                self.helper = Zeroconf.Avahi(*args, **kwargs)
            except ImportError:
                logger.warning('pybonjour nor avahi found, cannot announce presence')
                self.helper = None

    def publish(self, *args, **kwargs):
        if self.helper is not None:
            self.helper.publish(*args, **kwargs)

    def unpublish(self):
        if self.helper is not None:
            self.helper.unpublish()
